import itertools
import textwrap
import shutil
import functools
from ..lib import Program
from ..signature import Signature
from .doc import Doc

def _section_prefix(prog, x):
    for section in sorted(prog, key=len, reverse=True):
        if tuple(section[:len(x)]) == x:
            return x
    return tuple()

def _endpoints(prog, root_section):
    for section in sorted(prog):
        if section[:len(root_section)] == root_section:
            s = Signature(prog[section])

            signature = ''
            for x in s.positional:
                signature += ' %s' % x.name
            for x in s.keyword1:
                signature += ' [%s]' % x.name
            if s.var_positional:
                signature += ' [%s ...]' % s.var_positional.name
            yield section, signature[1:]

def _join(f):
    def decorator(*args, **kwargs):
        return '\n'.join(f(*args, **kwargs))
    return decorator

@_join
def _format_arg(prefix, indent, x):
    columns, _ = shutil.get_terminal_size((80, 20))
    argname, desc = x
    whitespace = ' ' * len(argname)
    n = columns - len(argname) - indent - len(prefix) - 2
    
    first = True
    for right in textwrap.wrap(desc, n):
        if first:
            left = argname + ': '
            first = False
        else:
            left = whitespace + '  '
        yield (' ' * indent) + left + right

def usage(name, f, h):
    prog = Program(f)
    section = _section_prefix(prog, h.section)
    p = {
        'name': name,
        'message': h.message,
        'endpoints': list(_endpoints(prog, section)),
    }
    if p['message']:
        yield 'error: %(message)s' % p
    for i, (section, signature) in enumerate(p['endpoints']):
        q = {
            'prefix': 'usage: ' if i == 0 else '       ',
            'name': p['name'],
            'sub': (' ' if section else '') + ' '.join(section),
            'signature': signature,
        }
        yield '%(prefix)s%(name)s%(sub)s [-help] [options] [--] %(signature)s' % q

def man(name, f, h):
    columns, _ = shutil.get_terminal_size((80, 20))

    prog = Program(f)
    section = _section_prefix(prog, h.section)
    g = prog.get(section)
    fdoc = Doc(g)
    p =  {
        'is_section': hasattr(g, 'items'),
        'is_callable': hasattr(g, '__call__'),
        'name': name,
        'endpoints': list(_endpoints(prog, section)),
        'description': fdoc.desc,
        'args': (_format_arg(' ', 2, a) for a in fdoc.args),
        'kwargs': (_format_arg('-', 2, a) for a in fdoc.kwargs),
    }

    yield 'SYNOPSIS'
    for section, signature in p['endpoints']:
        q = {
            'pointer': '> ' if p['is_section'] else '  ',
            'name': p['name'],
            'sub': ' '.join(section),
            'signature': signature,
        }
        yield '%(pointer)s %(name)s %(sub)s [-help] [options] [--] %(signature)s' % q
    if p['description']:
        yield 'DESCRIPTION'
        for line in textwrap.wrap(p['description'], columns-2):
            yield '  ' + line
    if p['is_callable']:
        yield 'INPUTS'
        for arg in p['args']:
            yield arg
    yield 'OPTIONS'
    for kwarg in p['kwargs']:
        yield kwarg
    try:
        kwarg
    except NameError:
        yield '  (None)'
    yield 'DETAIL'
    yield '  Run "-help" with a particular subcommand for more help.'



import logging
import os
import re
from collections import namedtuple
from functools import wraps
from ..lib import Program
from ..signature import Signature
from .doc import Doc

try:
    from ..templates import Jinja2
except ImportError:
    import sys
    def template(filename):
        def decorator(f):
            def wrapper(*args, **kwargs):
                logger.error('Jinja2 is not installed; run this: pip install horetu[all]')
                sys.exit(1)
        return decorator
else:
    _template_dir = os.path.abspath(os.path.join(__file__, '..', 'templates'))
    template = Jinja2(_template_dir).file

logger = logging.getLogger(__name__)

WebResponse = namedtuple('WebResponse', ['content_type', 'data'])
def content_type(x):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return WebResponse(x, func(*args, **kwargs))
        return wrapper
    return decorator

def _endpoints(prog, root_section):
    for section in sorted(prog):
        if section[:len(root_section)] == root_section:
            s = Signature(prog[section])

            signature = ''
            for x in section:
                signature += '/%s' % x
            for x in s.positional:
                signature += '/:%s' % x.name
            for x in s.keyword1:
                signature += '/[%s]' % x.name
            if s.var_positional:
                signature += '/&lt;%s&gt;' % s.var_positional.name
            signature += '?(option=...)'
            yield signature

@content_type('text/html')
@template('wsgi-usage.html')
def usage(f, h):
    prog = Program(f)
    if hasattr(prog.get(h.section), '__call__'):
        sections = []
        endpoints = list(_endpoints(prog, h.section))
    else:
        sections = ['/'.join(h.section + s) for s in prog]
        endpoints = []
    return {
        'message': h.message,
        'sections': sections,
        'endpoints': endpoints,
    }

@content_type('text/html')
@template('wsgi-man.html')
def man(f, h):
    prog = Program(f)
    g = prog.get(h.section)
    fdoc = Doc(g)
    return {
        'is_section': hasattr(g, 'items'),
        'is_callable': hasattr(g, '__call__'),
        'endpoints': list(_endpoints(prog, h.section)),
        'description': list(filter(None, re.split('r[\n\r]{2,}', fdoc.desc))),
        'args': fdoc.args,
        'kwargs': fdoc.kwargs,
    }



import os
from abc import ABCMeta
from functools import wraps, partial

def AppIter(x):
    '''
    :param x: Anything (returned from the underlying program)
    :rtype: Iterable of bytes
    :returns: Output as bytes
    '''
    if x is None:
        raise StopIteration
    elif isinstance(x, bytes):
        yield x
    elif isinstance(x, str):
        yield x.encode('utf-8')
    elif hasattr(x, '__iter__'):
        for y in x:
            yield from AppIter(y)
    else:
        yield format(x).encode('utf-8')

class TemplateEnvironment(metaclass=ABCMeta):
    raw_templates = {}
    def __init__(self, directory=None):
        if directory:
            self._directory = os.path.abspath(directory)
        else:
            self._directory = None

    @staticmethod
    def Template(x):
        raise NotImplementedError('You must define the Template method.')

    @property
    def render(self):
        msg = 'Set the %(c).render property to the %(c).Template method for rendering.'
        raise NotImplementedError(msg % {'c': self.__class__.__name__})

    def string(self, template):
        '''
        Render a template from a string.

        :param str template: The string of the template
        :rtype: decorator
        '''
        tpl = Template(path)
        return self._render(tpl)

    def file(self, filename):
        '''
        Render a template from a file.

        :param str filename: The file containing the template
        :rtype: decorator
        '''
        abspath = os.path.join(self._directory, filename)
        if abspath not in self.raw_templates:
            with open(abspath) as fp:
                self.raw_templates[abspath] = self.Template(fp.read())
        return self._render(self.raw_templates[abspath])

    def _render(self, tpl):
        def decorator(func):
            '''
            :param func: Function that returns a mapping (dict)
            '''
            @wraps(func)
            def wrapper(*args, **kwargs):
                fields = func(*args, **kwargs)
                if isinstance(self.render, str):
                    f = getattr(tpl, self.render)
                else:
                    f = partial(self.render, tpl)
                return AppIter(f(**fields))
            return wrapper
        return decorator

import string
class Template(TemplateEnvironment):
    Template = string.Template
    render = 'safe_substitute'

try:
    import airspeed
except ImportError:
    pass
else:
    class Airspeed(TemplateEnvironment):
        Template = airspeed.Template
        render = 'merge'

try:
    import bottle
except ImportError:
    pass
else:
    class Bottle(TemplateEnvironment):
        Template = bottle.SimpleTemplate
        render = 'render'

try:
    import django.conf, django.template
    django.conf.settings.configure()
except ImportError:
    pass
else:
    class Django(TemplateEnvironment):
        Template = django.template.Template
        @staticmethod
        def render(tpl, **ctx):
            return tpl.render(django.template.Context(ctx))

try:
    import jinja2
except ImportError:
    pass
else:
    class Jinja2(TemplateEnvironment):
        Template = jinja2.Template
        render = 'render'

try:
    import mako.template
except ImportError:
    pass
else:
    class Mako(TemplateEnvironment):
        Template = mako.template.Template
        render = 'render'

try:
    import moody
except ImportError:
    pass
else:
    class Moody(TemplateEnvironment):
        Template = moody.compile
        render = 'render'

try:
    import pystache
except ImportError:
    pass
else:
    class Mustache(TemplateEnvironment):
        Template = pystache.parse
        @staticmethod
        def render(tpl, **ctx):
            pystache.renderer

try:
    import pyratemp
except ImportError:
    pass
else:
    class Pyratemp(TemplateEnvironment):
        Template = pyratemp.Template
        render = '__call__'

try:
    import tempita
except ImportError:
    pass
else:
    class Tempita(TemplateEnvironment):
        Template = tempita.Template
        render = 'substitute'

try:
    import tonnikala.loader
except ImportError:
    pass
else:
    class Tonnikala(TemplateEnvironment):
        Template = tonnikala.loader.Loader().load_string
        render = 'render'

try:
    import trender
except ImportError:
    pass
else:
    class TRender(TemplateEnvironment):
        Template = trender.TRender
        render = 'render'

try:
    import wheezy.template
except ImportError:
    pass
else:
    class Wheezy(TemplateEnvironment):
        @staticmethod
        def Template(x):
            name = 'unnecessary-template-name'
            l = wheezy.template.DictLoader({name: x})
            e = wheezy.template.Engine(loader=l,
                                       extensions=[wheezy.template.CoreExtension()])
            return e.get_template(name)
        render = 'render'

'''
# Template template
try:
    import 
except ImportError:
    pass
else:
    class (TemplateEnvironment):
        Template = 
        render = ''
'''



import os
import logging
from io import StringIO
from collections import defaultdict
from pprint import pformat
from configparser import ConfigParser
from funmap import funmap

logger = logging.getLogger(__name__)

MAIN = 'dummy main section'

def Defaults(config_file, config_sep):
    '''
    :param str config_file: Configuration file
    :type config_sep: str or NoneType
    ;param config_sep: Character to separate sections, or None to indicate no sectioning
    :rtype: dict
    :returns: Stuff from the config file
    '''
    d = _rdict()

    if config_file and os.path.isfile(config_file):
        if config_sep:
            fp = open(config_file)
        else:
            fp = StringIO()
            fp.write('[%s]\n' % MAIN)
            with open(config_file) as gp:
                fp.write(gp.read())
            fp.seek(0)

        c = ConfigParser()
        c.read_file(fp)

        if config_sep:
            for config_section in c.sections():
                y = d
                if config_sep:
                    for component in config_section.split(config_sep):
                        y = y[component]
                y.update(c[config_section])

                w = (config_file, config_section, pformat(y))
                logger.info('Defaults from %s,\n  section [%s]:\n%s' % w)

            d = _as_dict(d)
        else:
            d = c[MAIN]

    else:
        d = {}
    return d

def _rdict():
    return defaultdict(_rdict)

def _as_dict(x):
    if isinstance(x, defaultdict):
        return {k: _as_dict(v) for k, v in x.items()}
    else:
        return x



import inspect
from enum import Enum
from funmap import funmap
from functools import partial
from ..lib import Program
from .defaults import Defaults
from ..signature import Signature
from ..templates import AppIter
from ..exceptions import (
    HoretuException, Error,
    HoretuShowHelp, HoretuCouldNotParse,
)

ANNOTATE_POS = 'Cannot annotate positional argument "%s"'

def Output(config_file, config_sep, function, inputs):
    '''
    :type config_file: str or None
    :param config_file: Location of the configuration file, or None
    :param str config_sep: Separator character between command and
        subcommand names in the configuration file (Default is " ".)
    :type function: dict, list, or callable
    :param function: The function or collection of functions
    :param Inputs inputs: Specification of the section,
        positional arguments, and keyword arguments
    :returns: Either the payload (iter of bytes) or a :py:class:`HoretuException`
    '''
    defaults = Defaults(config_file, config_sep)

    s = Signature(inputs.func)
    g = _with_defaults(inputs.func, defaults, s)

    fail = None
    args = reify_args(s, inputs.section, inputs.positionals)
    if isinstance(args, HoretuException):
        fail = args

    kwargs = reify_kwargs(s, inputs.flags)
    if isinstance(kwargs, HoretuShowHelp):
        fail = kwargs

    if fail == None:
        y = g(*args, **kwargs)
        return AppIter(y)
    else:
        fail.section = inputs.section
        raise fail

def reify_args(s, section, positionals):
    return _reify_args(s, positionals[max(len(section)-1,0):])

def _reify_args(s, str_args):
    '''
    :param Signature s:
    '''
    xs = iter(str_args)
    args = []
    fail = None

    vp = [s.var_positional] if s.var_positional else []
    for a in s.positional + s.keyword1 + vp:
        if not a.supports_positional:
            return TypeError(ANNOTATE_POS % a.name)

    for a in s.positional:
        try:
            x = next(xs)
        except StopIteration:
            fail = HoretuCouldNotParse(message='Not enough arguments')
        else:
            args.append(a.call(x))

    for a in s.keyword1:
        try:
            x = next(xs)
        except StopIteration:
            break
        else:
            args.append(a.call(x))

    a = s.var_positional
    if a:
        args.extend(map(a.call, xs))
    else:
        xs = list(xs)
        if xs:
            fail = HoretuCouldNotParse('Extra arguments -- %s' % ' '.join(xs))

    if fail == None:
        return args
    else:
        return fail

def reify_kwargs(s, str_kwargs):
    kwargs = {}
    for name, value in str_kwargs:
        a = s.keyword2.get(name)
        if a:
            try:
                kwargs[name] = a.call(value)
            except ValueError as e:
                return HoretuCouldNotParse(e.args[0])
            except HoretuException as e:
                return e
    return kwargs

def _with_defaults(f, defaults, sig):
    def w(*args, **kwargs):
        for key, value in defaults.items():
            if key in sig.keyword2 and key not in kwargs:
                kwargs[key] = sig.keyword2[key].call(value)
        return f(*args, **kwargs)
    return w



from functools import wraps

class CallableDict(dict):
    '''
    Use the new method, not the init method.
    '''
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.__name__ = func.__name__
        super(CallableDict, self).__init__(*args, **kwargs)
    def __call__(self, *args, **kwargs):
        self.func(*args, **kwargs) 
    def __copy__(self):
        return CallableDict.new(self.func, self)

    @classmethod
    def new(Class, func, dictionary):
        return wraps(func)(Class(func, dictionary))

def config_template(function):
    '''
    Write a configuration file with all the default parameter values.
    You can import this in your program to add a flag for creating a
    template configuration file. ::

        def main(x, y, z, sample_config=False):
            if sample_config:
                sys.stdout.write(horetu.config_template(main))
            else:
                ...

    Or, rather than including it in your end-user program, just call it
    once to create a template, and distribute that.
    '''
    raise NotImplementedError
